SystemOrganization addCategory: #'MethodMassage-Coverage'!
SystemOrganization addCategory: #'MethodMassage-Kernel'!
SystemOrganization addCategory: #'MethodMassage-Tests'!

Object subclass: #AssemblerAbsentClassImport
	instanceVariableNames: 'className'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!AssemblerAbsentClassImport class methodsFor: 'instance creation' stamp: 'eem 3/23/2017 14:09'!
className: nameOfAnAbsentClass
	^self new className: nameOfAnAbsentClass; yourself! !

!AssemblerAbsentClassImport methodsFor: 'accessing' stamp: 'eem 3/24/2017 10:38'!
binding
	^Association key: className value: self! !

!AssemblerAbsentClassImport methodsFor: 'accessing' stamp: 'eem 3/23/2017 14:08'!
className

	^className! !

!AssemblerAbsentClassImport methodsFor: 'accessing' stamp: 'eem 3/23/2017 13:35'!
className: anObject
	
	className := anObject.
! !

!AssemblerAbsentClassImport methodsFor: 'converting' stamp: 'eem 3/23/2017 14:10'!
forTheMetaclass
	^(className endsWith: ' class')
		ifTrue: [self]
		ifFalse: [self class className: className, ' class']! !

!AssemblerAbsentClassImport methodsFor: 'compiling' stamp: 'eem 3/24/2017 10:50'!
instVarNamesAndOffsetsDo: aBinaryBlock
	"This is part of the interface between the compiler and a class's instance or field names.
	 The class should enumerate aBinaryBlock with the instance variable name strings and
	 their integer offsets.  The order is important. Names evaluated later will override the
	 same names occurring earlier."

	"Nothing to do here; We assume Context et al are present."
	^self! !

!AssemblerAbsentClassImport methodsFor: 'printing' stamp: 'eem 3/24/2017 09:22'!
printOn: aStream
	super printOn: aStream.
	className ifNotNil:
		[aStream nextPutAll: ' className: '; store: className]! !

Object subclass: #AssemblerEncoderInterface
	instanceVariableNames: 'assembler encoder'
	classVariableNames: 'SpecialSelectors'
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!AssemblerEncoderInterface commentStamp: 'eem 2/22/2013 13:10' prior: 0!
An AssemblerEncoderInterface is an abstract superclass for the interfaces between a BytecodeAssembler and a BytecodeEncoder.  These interfaces convert the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) to the sizing and emitting messages implemented by BytecodeEncoder's subclasses.

Instance Variables
	assembler:		<BytecodeAssembler>
	encoder:		<BytecodeEncoder>!

!AssemblerEncoderInterface class methodsFor: 'instance creation' stamp: 'eem 2/22/2013 11:00'!
assembler: anAssembler encoder: anEncoder
	^self new assembler: anAssembler encoder: anEncoder! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 13:29'!
copySizeCodeToEmitterCode
	"AssemblerEncoderInterface copySizeCodeToEmitterCode"
	BytecodeSizer selectorsAndMethodsDo:
		[:s :m| | emitterMethod |
		((m literals includes: #shouldBeImplemented) not
		and: [(emitterMethod := BytecodeEmitter compiledMethodAt: s) literals includes: #shouldBeImplemented]) ifTrue:
			[BytecodeEmitter
				compile:
					(m getSourceFromFile asString
						copyReplaceAll: 'encoder size'
						with: 'encoder gen')
					classified: #assembly]]! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 11:31'!
createCodeSkeletons
	"BytecodeSizer createCodeSkeletons"
	"BytecodeEmitter createCodeSkeletons"
	InstructionClient selectorsAndMethodsDo:
		[:s :m|
		Parser new
			initPattern: m getSourceFromFile asString
			notifying: nil
			return:
				[:tuple|
				[:sel :args :precedence|
				self compile:
					(String streamContents:
						[:strm|
						precedence = 1
							ifTrue: [strm nextPutAll: sel]
							ifFalse:
								[sel keywords with: args do:
									[:kw :arg|
									strm nextPutAll: kw; space; nextPutAll: arg; space].
								 strm skip: -1].
						strm crtab; nextPutAll: 'self shouldBeImplemented'])
					classified: #assembly.
				] valueWithArguments: tuple]]! !

!AssemblerEncoderInterface class methodsFor: 'class initialization' stamp: 'eem 2/22/2013 13:20'!
initialize
	"AssemblerEncoderInterface initialize"
	SpecialSelectors := IdentityDictionary new.
	1 to: Smalltalk specialSelectorSize do:
		[:i | SpecialSelectors at: (Smalltalk specialSelectorAt: i) put: i]! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
assembler
	"Answer the value of assembler"

	^ assembler! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
assembler: anObject
	"Set the value of assembler"

	assembler := anObject! !

!AssemblerEncoderInterface methodsFor: 'initialize-release' stamp: 'eem 2/22/2013 11:01'!
assembler: anAssembler encoder: anEncoder
	assembler := anAssembler.
	encoder := anEncoder! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
encoder
	"Answer the value of encoder"

	^ encoder! !

!AssemblerEncoderInterface methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:59'!
encoder: anObject
	"Set the value of encoder"

	encoder := anObject! !

AssemblerEncoderInterface subclass: #BytecodeEmitter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeEmitter commentStamp: 'eem 2/22/2013 11:18' prior: 0!
A BytecodeEmitter converts the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) and sends them to the corresponding emitting messages implemented by BytecodeEncoder's subclasses.

Instance Variables: see superclass!

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 8/1/2014 07:13'!
blockReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder genReturnSpecialLiteralToCaller: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self blockReturnTop]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:27'!
blockReturnTop
	^encoder genReturnTopToCaller! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/24/2017 09:56'!
branchIfInstanceOf: behaviorOrArrayOfBehavior distance: distanceOrLabel
	| litIndex distance |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehavior.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehavior].
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^encoder genBranchIfInstanceOf: litIndex distance: distance! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/24/2017 09:55'!
branchIfInstanceOfLiteralAtIndex: litIndex distance: distanceOrLabel
	| distance |
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^encoder genBranchIfInstanceOf: litIndex distance: distance! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/24/2017 09:55'!
branchIfNotInstanceOf: behaviorOrArrayOfBehavior distance: distanceOrLabel
	| litIndex distance |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehavior.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehavior].
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^encoder genBranchIfNotInstanceOf: litIndex distance: distance! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/24/2017 09:55'!
branchIfNotInstanceOfLiteralAtIndex: litIndex distance: distanceOrLabel
	| distance |
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^encoder genBranchIfNotInstanceOf: litIndex distance: distance! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 4/10/2015 11:03'!
callInlinePrimitive: primitiveIndex
	^encoder genCallInlinePrimitive: primitiveIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:36'!
callPrimitive: primitiveIndex
	^encoder genCallPrimitive: primitiveIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 4/6/2015 12:08'!
directedSuperSend: selector numArgs: nArgs
	^encoder genSendDirectedSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
doDup
	^encoder genDup! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
doPop
	^encoder genPop! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:04'!
extA: anInteger 
	^encoder genUnsignedSingleExtendA: anInteger! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:04'!
extB: anInteger 
	^encoder genUnsignedSingleExtendB: anInteger! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:34'!
jump: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^encoder genJump: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:10'!
jump: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	condition
		ifTrue: [encoder genBranchPopTrue: offset]
		ifFalse: [encoder genBranchPopFalse: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:47'!
jumpLong: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^encoder genJumpLong: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:47'!
jumpLong: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^condition
		ifTrue: [encoder genBranchPopTrueLong: offset]
		ifFalse: [encoder genBranchPopFalseLong: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:28'!
methodReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder genReturnSpecialLiteral: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self methodReturnTop]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:27'!
methodReturnReceiver
	^encoder genReturnReceiver! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
methodReturnTop
	^encoder genReturnTop! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 18:03'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genStorePopLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:46'!
popIntoLiteralVariableAtIndex: litIndex
	^encoder genStorePopLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:42'!
popIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genStorePopInstVarLong: offset]
		ifFalse: [encoder genStorePopInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:44'!
popIntoRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"Variation of popIntoRemoteTemp:inVectorAt: that implies an immutability check."
	^encoder genStorePopRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:54'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genStorePopRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:09'!
popIntoTemporaryVariable: offset
	encoder genStorePopTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
pushActiveContext
	^encoder genPushThisContext! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/24/2015 15:56'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSizeOrLabel
	"N.B. If blockSizeOrLabel is a label it is expected be a label after the
	 last return in the block and before the bytecode following the block."
	| blockSize |
	blockSize := (assembler isLabel: blockSizeOrLabel)
					ifTrue: [assembler spanToBlockLabel: blockSizeOrLabel]
					ifFalse: [blockSizeOrLabel].
	^encoder genPushClosureCopyNumCopiedValues: numCopied numArgs: numArgs jumpSize: blockSize! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
pushConsArrayWithElements: numElements
	^encoder genPushConsArray: numElements! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:36'!
pushConstant: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		["rubbish way of handling special literals"
		 ([encoder genPushSpecialLiteral: aLiteral]
				on: Error
				do: [:ex| ex return: nil]) ifNotNil:
			[^self].
		 litIndex := encoder litIndex: aLiteral].
	^encoder genPushLiteral: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:40'!
pushConstantAtIndex: literalIndex
	^encoder genPushLiteral: literalIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/23/2017 16:42'!
pushFullClosure: compiledBlockLiteralIndex numCopied: numCopied
	^encoder genPushFullClosure: compiledBlockLiteralIndex numCopied: numCopied! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/23/2017 16:43'!
pushFullClosureAtIndex: compiledBlockLiteralIndex numCopied: numCopied receiverOnStack: rcvrOnStack ignoreOuterContext: ignoreOuterContext
	^encoder genPushFullClosure: compiledBlockLiteralIndex numCopied: numCopied receiverOnStack: rcvrOnStack ignoreOuterContext: ignoreOuterContext! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:10'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genPushLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
pushLiteralVariableAtIndex: litIndex
	^encoder genPushLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:50'!
pushNewArrayOfSize: numElements
	^encoder genPushNewArray: numElements! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
pushReceiver
	^encoder genPushReceiver! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:42'!
pushReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genPushInstVarLong: offset]
		ifFalse: [encoder genPushInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:45'!
pushRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"For symmetry with pop/storeIntoRemoteTemp:inVectorAt:."
	^encoder genPushRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genPushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:49'!
pushSpecialConstant: aLiteral 
	^encoder genPushSpecialLiteral: aLiteral! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:29'!
pushTemporaryVariable: offset
	^encoder genPushTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:31'!
send: selector super: supered numArgs: nArgs
	supered ifFalse:
		[(SpecialSelectors at: selector ifAbsent: []) ifNotNil:
			[:specialSelectorIndex|
			 ^encoder genSendSpecial: specialSelectorIndex numArgs: nArgs].
		^encoder genSend: (encoder sharableLitIndex: selector) numArgs: nArgs].
	^encoder genSendSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/22/2013 15:25'!
sendToAbsentDynamicSuperclass: selector numArgs: numArgs
	^encoder genSendAbsentDynamicSuper: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/15/2013 13:46'!
sendToAbsentImplicitReceiver: selector numArgs: numArgs
	^encoder genSendAbsentImplicit: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:40'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder genStoreLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:46'!
storeIntoLiteralVariableAtIndex: litIndex
	^encoder genStoreLiteralVar: litIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:41'!
storeIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder genStoreInstVarLong: offset]
		ifFalse: [encoder genStoreInstVar: offset]! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:45'!
storeIntoRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"Variation of storeIntoRemoteTemp:inVectorAt: that implies an immutability check."
	^encoder genStoreRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder genStoreRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:43'!
storeIntoTemporaryVariable: offset
	^encoder genStoreTemp: offset! !

!BytecodeEmitter methodsFor: 'assembly' stamp: 'eem 7/11/2016 13:48'!
trap
	^encoder genTrap! !

AssemblerEncoderInterface subclass: #BytecodeSizer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeSizer commentStamp: 'eem 2/22/2013 11:18' prior: 0!
A BytecodeSizer converts the assembly messages (see CompiledMethod>>abstractBytecodeMessages and InstructionStream interpretNextInstructionFor: et al) and sends them to the corresponding sizing messages implemented by BytecodeEncoder's subclasses.

Instance Variables: see superclass!

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 8/1/2014 07:14'!
blockReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder sizeReturnSpecialLiteralToCaller: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self blockReturnTop]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:26'!
blockReturnTop
	^encoder sizeReturnTopToCaller! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/24/2017 10:23'!
branchIfInstanceOf: behaviorOrArrayOfBehaviors distance: distanceOrLabel
	| litIndex distance |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehaviors.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehaviors].
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^distance ifNotNil:
		[encoder sizeBranchIfInstanceOf: litIndex distance: distance]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/24/2017 10:23'!
branchIfInstanceOfLiteralAtIndex: litIndex distance: distanceOrLabel
	| distance |
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^distance ifNotNil:
		[encoder sizeBranchIfInstanceOf: litIndex distance: distance]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/24/2017 10:23'!
branchIfNotInstanceOf: behaviorOrArrayOfBehaviors distance: distanceOrLabel
	| litIndex distance |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehaviors.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehaviors].
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^distance ifNotNil:
		[encoder sizeBranchIfNotInstanceOf: litIndex distance: distance]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/24/2017 10:23'!
branchIfNotInstanceOfLiteralAtIndex: litIndex distance: distanceOrLabel
	| distance |
	distance := (assembler isLabel: distanceOrLabel)
				ifTrue: [assembler spanToLabel: distanceOrLabel]
				ifFalse: [distanceOrLabel].
	^distance ifNotNil:
		[encoder sizeBranchIfNotInstanceOf: litIndex distance: distance]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 4/10/2015 11:04'!
callInlinePrimitive: primitiveIndex
	^encoder sizeCallInlinePrimitive: primitiveIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:35'!
callPrimitive: primitiveIndex
	^encoder sizeCallPrimitive: primitiveIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 4/6/2015 12:08'!
directedSuperSend: selector numArgs: nArgs
	^encoder sizeSendDirectedSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:46'!
doDup
	^encoder sizeDup! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:21'!
doPop
	^encoder sizePop! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:01'!
extA: anInteger 
	^encoder sizeUnsignedSingleExtendA: anInteger! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 14:00'!
extB: anInteger 
	^encoder sizeUnsignedSingleExtendB: anInteger! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:02'!
jump: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset
		ifNotNil: [encoder sizeJump: offset]
		ifNil: ["If this is a backward jump that is as-yet unsized, answer the long backward jump size.
			   c.f. e.g. MessageNode>>#sizeCodeForWhile:value:"
			  offsetOrLabel value <= assembler index ifTrue:
				[encoder sizeJumpLong: -1]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:21'!
jump: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[condition
			ifTrue: [encoder sizeBranchPopTrue: offset]
			ifFalse: [encoder sizeBranchPopFalse: offset]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:48'!
jumpLong: offsetOrLabel
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[encoder sizeJumpLong: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 15:38'!
jumpLong: offsetOrLabel if: condition
	| offset |
	offset := (assembler isLabel: offsetOrLabel)
				ifTrue: [assembler spanToLabel: offsetOrLabel]
				ifFalse: [offsetOrLabel].
	^offset ifNotNil:
		[condition
			ifTrue: [encoder sizeBranchPopTrueLong: offset]
			ifFalse: [encoder sizeBranchPopFalseLong: offset]]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:27'!
methodReturnConstant: aLiteral
	"Rubbish way of handling special constants."
	^[encoder sizeReturnSpecialLiteral: aLiteral]
		on: Error
		do: [:ex| ex return: (self pushLiteral: aLiteral) + self methodReturnTop]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 15:26'!
methodReturnReceiver
	^encoder sizeReturnReceiver! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:21'!
methodReturnTop
	^encoder sizeReturnTop! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 18:03'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizeStorePopLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
popIntoLiteralVariableAtIndex: litIndex
	^encoder sizeStorePopLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
popIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizeStorePopInstVarLong: offset]
		ifFalse: [encoder sizeStorePopInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:41'!
popIntoRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"Variation of popIntoRemoteTemp:inVectorAt: that implies an immutability check."
	^encoder sizeStorePopRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizePushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:14'!
popIntoTemporaryVariable: offset
	^encoder sizeStorePopTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:47'!
pushActiveContext
	^encoder sizePushThisContext! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/24/2015 15:51'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSizeOrLabel
	"N.B. If blockSizeOrLabel is a label it is expected be a label after the
	 last return in the block and before the bytecode following the block."
	| blockSize |
	blockSize := (assembler isLabel: blockSizeOrLabel)
					ifTrue: [assembler spanToBlockLabel: blockSizeOrLabel]
					ifFalse: [blockSizeOrLabel].
	^blockSize ifNotNil:
		[encoder sizePushClosureCopyNumCopiedValues: numCopied numArgs: numArgs jumpSize: blockSize]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:51'!
pushConsArrayWithElements: numElements
	^encoder sizePushConsArray: numElements! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:28'!
pushConstant: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		["rubbish way of handling special literals"
		 [(encoder sizePushSpecialLiteral: aLiteral) ifNotNil:
			[:size| ^size]]
			on: Error
			do: [:ex| ].
		 litIndex := encoder litIndex: aLiteral].
	^encoder sizePushLiteral: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:44'!
pushConstantAtIndex: literalIndex
	^encoder sizePushLiteral: literalIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'cb 4/10/2016 09:35'!
pushFullClosure: compiledBlockLiteralIndex numCopied: numCopied
	^encoder sizePushFullClosure: compiledBlockLiteralIndex numCopied: numCopied! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/23/2017 16:43'!
pushFullClosureAtIndex: compiledBlockLiteralIndex numCopied: numCopied receiverOnStack: rcvrOnStack ignoreOuterContext: ignoreOuterContext
	^encoder sizePushFullClosure: compiledBlockLiteralIndex numCopied: numCopied receiverOnStack: rcvrOnStack ignoreOuterContext: ignoreOuterContext! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:09'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizePushLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:47'!
pushLiteralVariableAtIndex: litIndex
	^encoder sizePushLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:48'!
pushNewArrayOfSize: numElements
	^encoder sizePushNewArray: numElements! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:20'!
pushReceiver
	^encoder sizePushReceiver! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
pushReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizePushInstVarLong: offset]
		ifFalse: [encoder sizePushInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:29'!
pushRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"For symmetry with pop/storeIntoRemoteTemp:inVectorAt:."
	^encoder sizePushRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:49'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizePushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 16:49'!
pushSpecialConstant: aLiteral 
	^encoder sizePushSpecialLiteral: aLiteral! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 12:50'!
pushTemporaryVariable: offset
	^encoder sizePushTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 13:16'!
send: selector super: supered numArgs: nArgs
	supered ifFalse:
		[(SpecialSelectors at: selector ifAbsent: []) ifNotNil:
			[:specialSelectorIndex|
			 ^encoder sizeSpecialSend: specialSelectorIndex numArgs: nArgs].
		^encoder sizeSend: (encoder sharableLitIndex: selector) numArgs: nArgs].
	^encoder sizeSendSuper: (encoder sharableLitIndex: selector) numArgs: nArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/22/2013 15:30'!
sendToAbsentDynamicSuperclass: selector numArgs: numArgs
	^encoder sizeSendAbsentDynamicSuper: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/15/2013 13:42'!
sendToAbsentImplicitReceiver: selector numArgs: numArgs
	^encoder sizeSendAbsentImplicit: (encoder sharableLitIndex: selector) numArgs: numArgs! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 21:40'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := encoder litIndexOrNilFor: aLiteral.
	litIndex ifNil:
		[litIndex := encoder litIndex: aLiteral].
	^encoder sizeStoreLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/14/2013 16:45'!
storeIntoLiteralVariableAtIndex: litIndex
	^encoder sizeStoreLiteralVar: litIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/24/2013 11:40'!
storeIntoReceiverVariable: offsetOrName
	| offset |
	offset := offsetOrName isInteger
				ifTrue: [offsetOrName]
				ifFalse: [assembler offsetForInstVarName: offsetOrName].
	"handle context inst var access"
	^(assembler shouldUseLongAccessForInstVarOffset: offset)
		ifTrue: [encoder sizeStoreInstVarLong: offset]
		ifFalse: [encoder sizeStoreInstVar: offset]! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 3/22/2017 16:41'!
storeIntoRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex
	"Variation of storeIntoRemoteTemp:inVectorAt: that implies an immutability check."
	^encoder sizeStoreRemoteInstanceVariable: tempIndex inObjectAt: InstVarIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/23/2013 17:48'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex
	^encoder sizeStoreRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 2/22/2013 16:42'!
storeIntoTemporaryVariable: offset
	^encoder sizeStoreTemp: offset! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 7/11/2016 13:48'!
trap
	^encoder sizeTrap! !

!BytecodeSizer methodsFor: 'assembly' stamp: 'eem 8/1/2014 20:42'!
trapIfNotInstanceOf: behaviorOrArrayOfBehavior
	| litIndex |
	litIndex := encoder litIndexOrNilFor: behaviorOrArrayOfBehavior.
	litIndex ifNil:
		[litIndex := encoder litIndex: behaviorOrArrayOfBehavior].
	^encoder sizeTrapIfNotInstanceOf: litIndex! !

Object subclass: #AssemblerMethod
	instanceVariableNames: 'literals instructions numArgs numTemps frameSize primitive flag signFlag noCountersFlag accessModifier trailer methodClass selector compiledMethodClass fixLabels encoder header initialPC endPC'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!AssemblerMethod commentStamp: 'eem 3/24/2017 08:26' prior: 0!
An AssemblerMethod is a holder for a sequence of literals and a sequence of instructions which can represent a disassembled method, or a method being assembled.  AssemblerMethod supports three modes of operation.

The first mode is transcription, CompiledMethod -> AssemblerMethod -> CompiledMethod.
e.g.
	| original "CompiledMethod"
	  assembler "AssemblerMethod"
	  replica "CompiledMethod" |
	original := Object >> #printOn:.
	assembler := original disassemble.
	replica := assembler assemble.
	self assert: original = replica

The second mode is to construct an AssemblerMethod and assemble it.  e.g.
	EncoderForSistaV1 assembler
		pushReceiver;
		trapIfNotInstanceOf: {SmallInteger. Character. LargePositiveInteger. LargeNegativeInteger.
							   Integer. Fraction. Magnitude};
		pushReceiver;
		methodReturnTop;
		assemble

	AssemblerMethod new
		methodClass: Object;
		selector: #printOn:;
		numArgs: 1;
		numTemps: 2;
		pushReceiver;
		send: #class;
		send: #name;
		popIntoTemporaryVariable: 1;
		pushTemporaryVariable: 0;
		doDup;
		pushTemporaryVariable: 1;
		send: #first;
		send: #isVowel;
		jump: 'L1' if: false;
		pushConstant: 'an ';
		jump: 'L2';
		label: 'L1';
		pushConstant: 'a ';
		label: 'L2';
		send: #nextPutAll:;
		doPop;
		pushTemporaryVariable: 1;
		send: #nextPutAll:;
		doPop;
		methodReturnReceiver;
		assemble

A third mode, useful for storage, inter-image transporation ,etc, is AssemblerMethod -> executable string -> AssemblerMethod.
e.g.
	| original "CompiledMethod"
	  assembler "AssemblerMethod"
	  assemblerText "String"
	  replica "CompiledMethod" |
	original := Object >> #printOn:.
	assembler := original disassemble.
	assemblerText := assembler assemblerString.
	replica := (Compiler evaluate: assemblerText) assemble.
	self assert: original = replica

Instance Variables
	accessModifier:			<nil|0,1,2,3>
	compiledMethodClass:	<nil|CompiledCode>
	encoder:				<BytecodeEncoder>
	endPC:					<nil|Integer>
	fixLabels:				<UndefinedObject|Boolean>
	flag:					<Boolean>
	frameSize:				<Integer>
	header:					<nil|Integer>
	initialPC:				<nil|Integer>
	instructions:			<SequenceableCollection>
	literals:					<SequenceableCollection>
	methodClass:			<Behavior>
	noCountersFlag:		<nil|Boolean>
	numArgs:				<Integer>
	numTemps:			<Integer>
	primitive:				<Integer>
	signFlag:				<Boolean>
	trailer:					<CompiledMethodTrailer>

accessModifier
	- if not nil this is the values of the two bits of flags that specify Newspeak access control in the method header

compiledMethodClass
	- if not nil this is the class to use to create a method from this assembly

encoder
	- the encoder to use to encode literals during assembly

endPC
	- this is the expected endPC of the resulting method.  If not nil it will be checked against the actual endPC

fixLabels
	- if nil, the AssemblerMethod is being used in transcription mode and branch offsets should be integers.
	  if a boolean then it indicates whether instructions contains one or more instructions that need their branches fixing up.

flag
	- the value of the method's header flag (see CompiledMethod>>flag)

frameSize
	- the size of the method's frame (see CompiledMethod>>frameSize)

header
	- this is the expected header of the resulting method.  If not nil it will be checked against the actual header

initialPC
	- this is the expected initialPC of the resulting method.  If not nil it will be checked against the actual initialPC

instructions
	- the sequence of Message and LookupKey objects making up the method's instructions

literals
	- the sequence of objects making up the method's literals

methodClass
	- the target class for the method (see CompiledMethod>>methodClass)

noCountersFlag
	- if not nil this is the value of the Sista JIT-without-performance-counters flag

numArgs
	- the method's argument count (see CompiledMethod>>numArgs)

numTemps
	- the method's temporary count (see CompiledMethod>>numTemps)

primitive
	- the method's primitive (see CompiledMethod>>primitive)

signFlag
	- the value of the method's signFlag (see CompiledMethod>>signFlag)

trailer
	- the method's trailer (see CompiledMethod>>trailer)

!

!AssemblerMethod class methodsFor: 'instance creation' stamp: 'eem 3/23/2017 14:52'!
literals: literalSequence instructions: instructionSequence numArgs: nArgs numTemps: nTemps frameSize: frameSizeNum primitive: primitiveIndex flag: flagBoolean signFlag: signFlagBoolean accessModifier: accessModifierBits noCounters: noCountersBoolean trailer: methodTrailer methodClass: class selector: aSelector compiledMethodClass: classOfCompiledMethod
	^self new
		literals: literalSequence
		instructions: instructionSequence
		numArgs: nArgs
		numTemps: nTemps
		frameSize: frameSizeNum
		primitive: primitiveIndex
		flag: flagBoolean
		signFlag: signFlagBoolean
		accessModifier: accessModifierBits
		noCounters: noCountersBoolean
		trailer: methodTrailer
		methodClass: class
		selector: aSelector
		compiledMethodClass: classOfCompiledMethod! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/23/2017 15:03'!
accessModifier
	
	^ accessModifier
! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/23/2017 15:03'!
accessModifier: anObject
	
	accessModifier := anObject.
! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 8/7/2014 11:10'!
assemble
	| assembler |
	assembler := BytecodeAssembler new.
	encoder ifNotNil: [assembler encoder: encoder].
	^assembler assemble: self! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/15/2013 13:18'!
assemblerString
	^String streamContents: [:s| self printAsAssemblerOn: s]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 4/26/2016 09:19'!
binding: aSymbol
	"For now, if the binding is not found, answer a fake binding rather than enter into Undeclared."
	self literal: ((methodClass bindingOf: aSymbol) ifNil:
					[Transcript cr; print: 'warning: ', aSymbol, ' is unbound'.
					 aSymbol -> nil])! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/23/2017 17:11'!
branchIfInstanceOf: behaviorOrArrayOfBehavior distance: label
	self checkLabel: label.
	instructions addLast: (Message
								selector: #branchIfInstanceOf:distance:
								arguments: {behaviorOrArrayOfBehavior. label})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/24/2017 09:53'!
branchIfInstanceOfLiteralAtIndex: litIndex distance: label
	self checkLabel: label.
	instructions addLast: (Message
								selector: #branchIfInstanceOfLiteralAtIndex:distance:
								arguments: {litIndex. label})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/23/2017 17:11'!
branchIfNotInstanceOf: behaviorOrArrayOfBehavior distance: label
	self checkLabel: label.
	instructions addLast: (Message
								selector: #branchIfNotInstanceOf:distance:
								arguments: {behaviorOrArrayOfBehavior. label})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/24/2017 09:53'!
branchIfNotInstanceOfLiteralAtIndex: litIndex distance: label
	self checkLabel: label.
	instructions addLast: (Message
								selector: #branchIfNotInstanceOfLiteralAtIndex:distance:
								arguments: {litIndex. label})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 7/12/2016 13:35'!
callInlinePrimitive: primNumber
	instructions addLast: (Message
								selector: #callPrimitive:
								argument: primNumber + 32768)! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/15/2013 16:06'!
checkLabel: label
	fixLabels ifNotNil:
		[label isInteger
			ifTrue: [self error: 'one cannot use integers as labels in assembler mode. integers are interpreted as branch distances']
			ifFalse: [fixLabels := true]]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 14:06'!
compiledMethodClass
	"Answer the value of compiledMethodClass"

	^ compiledMethodClass! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 14:06'!
compiledMethodClass: anObject
	"Set the value of compiledMethodClass"

	compiledMethodClass := anObject! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/6/2015 11:56'!
directedSuperSend: aSelector
	self directedSuperSend: aSelector numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/24/2015 15:17'!
doesNotUnderstand: aMessage
	(BytecodeSizer includesSelector: aMessage selector) ifFalse:
		[^super doesNotUnderstand: aMessage].
	aMessage lookupClass: nil.
	instructions addLast: aMessage! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
dup
	"For compatibility with InstructionPrinter output"
	^self doDup! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:07'!
encoder
	^encoder! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:08'!
encoder: anObject
	encoder := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 08:30'!
endPC
	
	^ endPC
! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 09:24'!
endPC: anObject

	endPC := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
extA: anInteger 
	instructions addLast: (Message selector: #extA: argument: anInteger)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
extB: anInteger 
	instructions addLast: (Message selector: #extB: argument: anInteger)! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/24/2017 10:29'!
fixLabelsIfNecessary
	| map |
	fixLabels ~~ true ifTrue: [^self].
	map := Dictionary new.
	instructions do:
		[:inst|
		(self isLabel: inst) ifTrue:
			[map at: inst key put: inst]].
	instructions do:
		[:inst|
		((self isLabel: inst) not
		 and: [#(jump: if: blockSize: distance:) includes: inst selector keywords last]) ifTrue:
			[inst selector first = $j
				ifTrue:
					[inst arguments first isInteger ifFalse:
						[inst arguments at: 1 put: (map at: inst arguments first)]]
				ifFalse:
					[inst arguments last isInteger ifFalse:
						[inst arguments at: inst numArgs put: (map at: inst arguments last)]]]].
	fixLabels := false! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
flag
	"Answer the value of flag"

	^ flag! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/14/2013 14:39'!
flag: anObject
	"Set the value of flag"

	flag := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
frameSize
	"Answer the value of frameSize"

	^ frameSize! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
frameSize: anObject
	"Set the value of frameSize"

	frameSize := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 08:30'!
header
	
	^ header
! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 09:24'!
header: anObject

	header := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/23/2017 14:21'!
importLiteral: classNameOrArrayOfClassNames
	"Add as a literal the resolution of classNameOrArrayOfClassNames to class(es) present in the system or AbsentClassImport(s) as appropriate."
	^self literal: (self resolveImportLiteral: classNameOrArrayOfClassNames)! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 09:18'!
importMethodClass
	"Answer the name of methodClass."

	^methodClass ifNotNil: [methodClass name]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/24/2017 09:03'!
importMethodClass: className
	"Set the methodClass to the resolution of className to the class present in the system or an AbsentClassImport as appropriate."
	^self methodClass: (self resolveImportLiteral: className)! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/24/2017 09:11'!
importStringFor: behaviorOrArrayOfBehaviors
	^behaviorOrArrayOfBehaviors isBehavior
		ifTrue: [behaviorOrArrayOfBehaviors name storeString]
		ifFalse: [(behaviorOrArrayOfBehaviors collect: [:behavior| behavior name]) printString]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 08:30'!
initialPC
	
	^ initialPC
! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 09:25'!
initialPC: anObject

	initialPC := anObject! !

!AssemblerMethod methodsFor: 'initialize-release' stamp: 'eem 8/1/2014 20:50'!
initialize
	fixLabels := false.
	compiledMethodClass := CompiledMethod.
	instructions := OrderedCollection new.
	literals := OrderedCollection new.
	frameSize := thisContext method frameSize. "should be small"
	flag := signFlag := false.
	numArgs := numTemps := primitive := 0! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:43'!
instructions
	"Answer the value of instructions"
	self fixLabelsIfNecessary.
	^instructions! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
instructions: anObject
	"Set the value of instructions"

	instructions := anObject! !

!AssemblerMethod methodsFor: 'private' stamp: 'eem 3/15/2013 10:26'!
isLabel: aMessageOrLookupKey
	^aMessageOrLookupKey isVariableBinding! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jump: label
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jump:
										argument: label)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jump: label if: cond
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jump:if:
										arguments: {label. cond})! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
jumpFalse: label
	"For compatibility with InstructionPrinter output"
	^self jump: label if: false! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jumpLong: label
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jumpLong:
										argument: label)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
jumpLong: label if: cond
	self checkLabel: label.
	instructions addLast: (Message
										selector: #jumpLong:if:
										arguments: {label. cond})! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
jumpTrue: label
	"For compatibility with InstructionPrinter output"
	^self jump: label if: true! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
label: aString 
	instructions addLast: aString -> nil! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/20/2013 12:56'!
labels
	^instructions select: [:ea| ea isVariableBinding]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 8/7/2014 11:23'!
literal: aLiteral
	literals addLast: aLiteral! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
literals
	"Answer the value of literals"

	^ literals! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
literals: anObject
	"Set the value of literals"

	literals := anObject! !

!AssemblerMethod methodsFor: 'initialize-release' stamp: 'eem 3/23/2017 14:48'!
literals: literalSequence instructions: instructionSequence numArgs: nArgs numTemps: nTemps frameSize: frameSizeNum primitive: primitiveIndex flag: flagBoolean signFlag: signFlagBoolean accessModifier: accessModifierBits noCounters: noCountersBoolean trailer: methodTrailer methodClass: class selector: aSelector compiledMethodClass: classOfCompiledMethod
	literals := literalSequence.
	instructions := instructionSequence.
	numArgs := nArgs.
	numTemps := nTemps.
	frameSize := frameSizeNum.
	primitive := primitiveIndex.
	flag := flagBoolean.
	signFlag := signFlagBoolean.
	noCountersFlag := noCountersBoolean.
	accessModifier := accessModifierBits.
	trailer := methodTrailer.
	methodClass := class.
	selector := aSelector.
	compiledMethodClass := classOfCompiledMethod! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/24/2017 08:19'!
methodClass
	"Answer the value of methodClass. Default to Object if nil (and if we're compiling a CompiledMehtod, not a CompiledBlock)"

	^methodClass ifNil:
		[(compiledMethodClass notNil
		  and: [compiledMethodClass includesBehavior: CompiledMethod]) ifTrue:
			[Object]]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:06'!
methodClass: anObject
	"Set the value of methodClass"

	methodClass := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/23/2017 17:22'!
noCountersFlag

	^noCountersFlag == true "nil & false => false"! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/23/2017 14:40'!
noCountersFlag: anObject
	
	noCountersFlag := anObject.
! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numArgs
	"Answer the value of numArgs"

	^ numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numArgs: anObject
	"Set the value of numArgs"

	numArgs := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numTemps
	"Answer the value of numTemps"

	^ numTemps! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
numTemps: anObject
	"Set the value of numTemps"

	numTemps := anObject! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
pop
	"For compatibility with InstructionPrinter output"
	^self doPop! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
popIntoLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #popIntoLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:59'!
popIntoTemp: n
	"For compatibility with InstructionPrinter output"
	^self popIntoTemporaryVariable: n! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
primitive
	"Answer the value of primitive"

	^ primitive! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/14/2013 14:38'!
primitive: anObject
	"Set the value of primitive"

	primitive := anObject! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/24/2017 09:16'!
printAccessorsForAssemblyOn: aStream
	| dummy |
	dummy := self class new.
	"Generate all the accessors; for now don't handle the trailer"
	#(			compiledMethodClass: importMethodClass: selector:
				primitive: numArgs: numTemps: frameSize:
				flag: signFlag: noCountersFlag: accessModifier:
				trailerData:)
		with: #(compiledMethodClass importMethodClass selector
				primitive numArgs numTemps frameSize
				flag signFlag noCountersFlag accessModifier
				trailerData)
		do: [:setter :getter|
			(self perform: getter) ifNotNil:
				[:val|
				val ~= (dummy perform: getter) ifTrue:
					[aStream crtab; nextPutAll: setter; space.
					 val isSymbol ifTrue: [aStream store: val] ifFalse: [aStream print: val].
					 aStream nextPut: $;]]]! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/24/2017 08:31'!
printAsAssemblerOn: aStream
	aStream nextPut: $(; print: self class; nextPutAll: ' new'.

	self printAccessorsForAssemblyOn: aStream.
	self printVerifiersForAssemblyOn: aStream.
	self printLiteralsForAssemblyOn: aStream.
	self printInstructionsForAssemblyOn: aStream.

	aStream crtab; nextPutAll: #yourself; nextPut: $)! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/24/2017 10:04'!
printInstructionsForAssemblyOn: aStream
	instructions do:
		[:messageOrLabel|
		aStream crtab.
		(self isLabel: messageOrLabel)
			ifTrue:
				[aStream nextPutAll: 'label: '; print: messageOrLabel key]
			ifFalse:
				[messageOrLabel numArgs = 0
					ifTrue: [aStream nextPutAll: messageOrLabel selector]
					ifFalse:
						[messageOrLabel selector keywords with: messageOrLabel arguments do:
							[:kwd :arg|
							aStream nextPutAll: kwd; space.
							aStream print: ((kwd = 'jump:' or: [kwd = 'blockSize:' or: [kwd = 'distance:']])
												ifTrue: [self assert: (self isLabel: arg).
														arg key]
												ifFalse: [arg]); space].
						aStream skip: -1]].
		aStream nextPut: $;]! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/23/2017 12:03'!
printLiteralsForAssemblyOn: aStream
	| literalsToDefine |
	literalsToDefine := literals copy asOrderedCollection.
	(literalsToDefine last isVariableBinding
	 and: [literalsToDefine last value = methodClass]) ifTrue:
		[literalsToDefine removeLast.
		 self assert: (literalsToDefine last == selector
					  or: [literalsToDefine last isMethodProperties
						  and: [literalsToDefine last selector == selector]]).
		 literalsToDefine removeLast].
	literalsToDefine do:
		[:litOrBinding|
		aStream crtab.
		litOrBinding isLiteral
			ifTrue: [aStream nextPutAll: 'literal: '; print: litOrBinding]
			ifFalse:
				[litOrBinding isVariableBinding 
					ifTrue: [aStream nextPutAll: 'binding: '; store: litOrBinding key] 
					ifFalse:
						[litOrBinding isCompiledCode
							ifTrue: [aStream nextPutAll: 'literal: '; print: nil "a post-process pass will fix this"]
							ifFalse:
								[(litOrBinding isBehavior or: [litOrBinding isArray  and: [litOrBinding allSatisfy: [:b | b isBehavior]]])
									ifTrue: [aStream nextPutAll: 'importLiteral: '; nextPutAll: (self importStringFor: litOrBinding)]
									ifFalse:
										[aStream nextPutAll: 'literal: '; store: litOrBinding ]]]]. "N.B. This handles nested AssemblerMethods for nested CompiledBlocks."
		aStream nextPut: $;]! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/24/2017 09:00'!
printVerifiersForAssemblyOn: aStream
	#(			header: endPC: initialPC:)
		with: #(header endPC initialPC)
		do: [:setter :getter|
			(self perform: getter) ifNotNil:
				[:val|
				aStream crtab; nextPutAll: setter; space.
				val storeOn: aStream base: (getter == #header ifTrue: [16] ifFalse: [10]).
				aStream nextPut: $;]]! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
push: aLiteral
	"Convenience for pushLiteral:.  Won't work unless an encoder has been set."
	"For compatibility with InstructionPrinter output"
	^(encoder isSpecialLiteralForPush: aLiteral)
		ifTrue: [self pushSpecialConstant: aLiteral]
		ifFalse: [self pushConstant: aLiteral]! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/24/2015 15:56'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	"N.B. If blockSizeOrLabel is a label it is expected be a label after the
	 last return in the block and before the bytecode following the block."
	self checkLabel: blockSize.
	instructions addLast: (Message
										selector: #pushClosureCopyNumCopiedValues:numArgs:blockSize:
										arguments: {numCopied. numArgs. blockSize})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushConstantAtIndex: litIndex
	instructions addLast: (Message
										selector: #pushConstantAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 3/23/2017 16:52'!
pushFullClosureAtIndex: litIndex numCopied: numCopied
	instructions addLast: (Message
								selector: #pushFullClosureAtIndex:numCopied:
								arguments: {litIndex. numCopied})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/23/2017 16:40'!
pushFullClosureAtIndex: litIndex numCopied: numCopied receiverOnStack: receiverOnStack ignoreOuterContext: ignoreOuterContext
	instructions addLast: (Message
								selector: #pushFullClosureAtIndex:numCopied:receiverOnStack:ignoreOuterContext:
								arguments: {litIndex. numCopied. receiverOnStack. ignoreOuterContext})! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #pushLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
pushSpecialConstant: aConstant
	instructions addLast: (Message
										selector: #pushSpecialConstant:
										argument: aConstant)! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:58'!
pushTemp: n
	"For compatibility with InstructionPrinter output"
	^self pushTemporaryVariable: n! !

!AssemblerMethod methodsFor: 'initialize-release' stamp: 'eem 3/24/2017 08:30'!
recordVerificationInfoFrom: aMethod
	"Record header, initialPC and endPC from aMethod for checking."
	header := aMethod header.
	endPC := aMethod endPC.
	initialPC := aMethod initialPC! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 3/23/2017 15:01'!
resolveImportLiteral: classNameOrArrayOfClassNames
	"Resolve classNameOrArrayOfClassNames to class(es) present in the system or AssemblerAbsentClassImport(s) as appropriate."
	classNameOrArrayOfClassNames isArray ifTrue:
		[^classNameOrArrayOfClassNames collect:
			[:className|
			self resolveImportLiteral: className]].
	(classNameOrArrayOfClassNames last == $s
	 and: [classNameOrArrayOfClassNames endsWith: ' class']) ifTrue:
		[| import |
		 import := self resolveImportLiteral: (classNameOrArrayOfClassNames allButLast: 6).
		 ^import isBehavior
			ifTrue: [import class]
			ifFalse: [import forTheMetaclass]].
	^(Smalltalk classNamed: classNameOrArrayOfClassNames)
		ifNotNil: [:aClass| aClass]
		ifNil: [AssemblerAbsentClassImport className: classNameOrArrayOfClassNames]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/24/2013 13:16'!
selector
	"Answer the value of selector"

	^ selector! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/24/2013 13:16'!
selector: anObject
	"Set the value of selector"

	selector := anObject! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:59'!
send: aSelector
	"For compatibility with InstructionPrinter output"
	self send: aSelector super: false numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 3/24/2015 15:11'!
sendToAbsentImplicitReceiver: aSelector
	self sendToAbsentImplicitReceiver: aSelector numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
signFlag
	"Answer the value of signFlag"

	^ signFlag! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 04:31'!
signFlag: anObject
	"Set the value of signFlag"

	signFlag := anObject! !

!AssemblerMethod methodsFor: 'assembling' stamp: 'eem 08/07/2014 11:24'!
storeIntoLiteralVariableAtIndex: litIndex
	instructions addLast: (Message
										selector: #storeIntoLiteralVariableAtIndex:
										argument: litIndex)! !

!AssemblerMethod methodsFor: 'printing' stamp: 'eem 3/23/2017 11:59'!
storeOn: aStream
	"Store the receiver by using printAsAssemblerOn:.
	 N.B. This is relied upon in the printLiteralsForAssemblyOn: method to deal with nested blocks."
	self printAsAssemblerOn: aStream! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/14/2015 09:59'!
superSend: aSelector
	"For compatibility with InstructionPrinter output"
	self send: aSelector super: true numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'assembling-shorthand' stamp: 'eem 4/6/2015 11:57'!
superSend: aSelector above: classOrBinding
	self pushLiteral: (classOrBinding isVariableBinding
						ifTrue: [classOrBinding]
						ifFalse: [classOrBinding binding]).
	self directedSuperSend: aSelector numArgs: aSelector numArgs! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 8/7/2014 11:08'!
trailer
	^trailer ifNil: [CompiledMethodTrailer empty]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 2/22/2013 10:06'!
trailer: anObject
	"Set the value of trailer"

	trailer := anObject! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:07'!
trailerData
	^trailer ifNotNil: [trailer rawData]! !

!AssemblerMethod methodsFor: 'accessing' stamp: 'eem 3/15/2013 10:07'!
trailerData: aByteArray
	self trailer: (CompiledMethodTrailer new method: aByteArray)! !

Object subclass: #BytecodeAssembler
	instanceVariableNames: 'code sizes encoder method index shouldNegateInstVarOffsetsMask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

!BytecodeAssembler commentStamp: 'eem 2/22/2013 11:12' prior: 0!
A BytecodeAssembler is something that converts an AssemblerMethod into a CompiledMethod.

Instance Variables
	code:		<AssemblerMethod>
	encoder:	<BytecodeEncoder>
	sizes:		<SequenceableCollection>

code
	- the input assembly method

encoder
	- the encoder that generates instructions

sizes
	- the size of each corresponding bytecode in code's instructions!

!BytecodeAssembler methodsFor: 'assembly-public' stamp: 'eem 3/24/2017 10:48'!
assemble: ass "<AssemblerMethod>"
	code := ass.
	sizes := Array new: code instructions size.
	self createEncoder: ass.
	self nilLabelTargets.
	self sizePasses.
	self createMethod.
	self emitPass.
	self verify: ass.
	^method! !

!BytecodeAssembler methodsFor: 'private' stamp: 'eem 3/24/2017 08:52'!
assert: expected equals: actual entity: entityName
	expected ~= actual ifTrue:
		[self error: 'method ', entityName, ' should be ', (entityName = 'header' ifTrue: [expected hex] ifFalse: [expected printString]),
					' but is ', (entityName = 'header' ifTrue: [actual hex] ifFalse: [actual printString])]! !

!BytecodeAssembler methodsFor: 'private' stamp: 'eem 3/24/2017 08:56'!
assert: expected or: anticipated equals: actual entity: entityName
	(expected ~= actual
	 and: [anticipated ~= actual]) ifTrue:
		[^self
			assert: ((expected - actual) abs <= (anticipated - actual) abs
						ifTrue: [expected]
						ifFalse: [anticipated])
			equals: actual
			entity: entityName]! !

!BytecodeAssembler methodsFor: 'accessing' stamp: 'eem 2/23/2013 16:53'!
code
	^code! !

!BytecodeAssembler methodsFor: 'instance initialization' stamp: 'cb 9/5/2016 09:56'!
createEncoder
	"The signFlag defines the bytecode set of a method.
	 So create a dummy method with its sign flag to that of
	 code's, and instantiate the dummy method's encoderClass."
	| dummyMethod newLits |
	encoder ifNil:
		[dummyMethod := CompiledMethod
							newMethod: 0
							header: (code signFlag ifTrue: [SmallInteger minVal] ifFalse: [0]).
		 encoder := dummyMethod encoderClass new].
	encoder
		initClass: code methodClass
		literalStream: ((newLits := self getLiterals: code literals) isEmpty ifFalse:
						[self
							ensureLiteralsIncludeMethodClass: newLits
							isQuick: (self isQuickPrimitiveIndex: code primitive)]).
	^encoder! !

!BytecodeAssembler methodsFor: 'instance initialization' stamp: 'eem 3/24/2017 10:48'!
createEncoder: assemblerMethod
	"The signFlag defines the bytecode set of a method.
	 So create a dummy method with its sign flag to that of
	 code's, and instantiate the dummy method's encoderClass."
	| dummyMethod newLits |
	encoder ifNil:
		[dummyMethod := CompiledMethod
							newMethod: 0
							header: (code signFlag ifTrue: [SmallInteger minVal] ifFalse: [0]).
		 encoder := dummyMethod encoderClass new].
	encoder
		initClass: code methodClass
		literalStream: ((newLits := self getLiterals: code literals) isEmpty ifFalse:
						[(assemblerMethod compiledMethodClass includesBehavior: CompiledMethod)
							ifTrue:
								[self
									ensureLiteralsIncludeMethodClass: newLits
									isQuick: (self isQuickPrimitiveIndex: code primitive)]
							ifFalse:
								[newLits]]).
	^encoder! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 3/23/2017 16:00'!
createMethod
	| header size |
	header := encoder
					computeMethodHeaderForNumArgs: code numArgs
					numTemps: code numTemps
					numLits: encoder allLiterals size
					primitive: code primitive.

	code flag ifTrue:
		[header := header bitOr: 1 << 29].
	code accessModifier ifNotNil:
		[:bits| header := header bitOr: bits << 28].
	code noCountersFlag ifTrue:
		[header := header bitOr: 1 << 15].

	size := sizes size = 0 ifTrue: [0] ifFalse: [self totalSizeFrom: 1 to: sizes size].
	method := code trailer
					createMethod: size
					class: code compiledMethodClass
					header: header.
	method needsFrameSize: code frameSize - method numTemps! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 3/24/2017 11:04'!
emitPass
	| literals nLits emitter |

	literals := encoder allLiterals.
	nLits := literals size.
	emitter := BytecodeEmitter assembler: self encoder: encoder.

	1 to: nLits do: [:lit | method literalAt: lit put: (literals at: lit)].

	encoder streamToMethod: method.

	code instructions withIndexDo:
		[:messageOrLabel :i|
		(self isLabel: messageOrLabel) ifFalse:
			[index := i.
			 messageOrLabel sentTo: emitter]].

	"Need to create distinct literals for all literals in the method
	 *except* variable bindings, the methodClass literal, and any exotica such as Behaviors and CompiledBlocks.
	 Need to hook nested CompiledBlocks up to their outer method."
	1 to: method numLiterals - 2 do:
		[:litIndex| | lit |
		lit := method literalAt: litIndex.
		lit isLiteral
			ifTrue:
				[method
					literalAt: litIndex
					put: (lit isArray ifTrue: [lit deepCopy] ifFalse: [lit copy])]
			ifFalse:
				[lit isCompiledCode ifTrue:
					[lit outerCode: method]]].

	"Need to create a distinct additional method state if it has one.
	 Avoid setting selector if already set (method penultimateLiteral notNil and not a metod properties)"
	method isCompiledMethod ifTrue:
		[method penultimateLiteral
			ifNil: [method selector: code selector]
			ifNotNil:
				[:plit|
				plit isMethodProperties ifTrue:
					[method
						penultimateLiteral: plit copy;
						selector: code selector]]]
	! !

!BytecodeAssembler methodsFor: 'assembly-public' stamp: 'eem 8/1/2014 06:05'!
encoder: aBytecodeEncoder
	encoder := aBytecodeEncoder! !

!BytecodeAssembler methodsFor: 'private' stamp: 'eem 3/21/2013 06:46'!
ensureLiteralsIncludeMethodClass: literals isQuick: isQuick
	"Check whether literals includes the methodClass association (and hence by implication the penultimate literal).
	 If not, add a suitable one.  In the assembly case the literal will be missing."
	literals size > 0 ifTrue:
		[| last |
		isQuick ifTrue: "support Newspeak accessors"
			[self assert: literals size = 2.
			 ^literals].
		last := literals last.
		(last isVariableBinding
		and: [last value = code methodClass
		and: [last key == nil or: [last == code methodClass binding]]]) ifTrue:
			[^literals]].
	^literals, { code selector. code methodClass ifNotNil: [:class| class binding] }! !

!BytecodeAssembler methodsFor: 'instance initialization' stamp: 'cb 9/5/2016 09:39'!
getLiterals: literals
	| newLits |
	newLits := Array new: literals size.
	literals withIndexDo: [:lit :i |
		newLits at: i put: (lit class == AssemblerMethod 
			ifTrue: [lit assemble]
			ifFalse: [lit])].
	^ newLits! !

!BytecodeAssembler methodsFor: 'accessing' stamp: 'eem 2/22/2013 15:01'!
index
	"For the sizer."
	^index! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/22/2013 12:47'!
isLabel: aMessageOrLookupKey
	^aMessageOrLookupKey isVariableBinding! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/23/2013 15:39'!
isQuickPrimitiveIndex: anInteger
	^anInteger between: 256 and: 519! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:29'!
nilLabelTargets
	code instructions do:
		[:messageOrLabel|
		(self isLabel: messageOrLabel) ifTrue:
			[messageOrLabel value: nil]]! !

!BytecodeAssembler methodsFor: 'testing' stamp: 'eem 2/24/2013 12:02'!
shouldUseLongAccessForInstVarOffset: offset
	"compute the mask of which inst var offsets need to be negated.
	 This is true of inst vars of contexts only."
	shouldNegateInstVarOffsetsMask ifNotNil:
		[:snivo| ^(snivo bitAnd: (1 bitShift: offset)) ~= 0].
	"The order in which this is done means that if the first offset is +ve all others will be too."
	shouldNegateInstVarOffsetsMask := 0.
	code methodClass instVarNamesAndOffsetsDo:
		[:n :o|
		o >= 0 ifTrue:
			[^(shouldNegateInstVarOffsetsMask bitAnd: (1 bitShift: offset)) ~= 0].
		o < 0 ifTrue:
			[shouldNegateInstVarOffsetsMask := shouldNegateInstVarOffsetsMask + (1 bitShift: -1 - o)]].
	^true! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/24/2013 13:53'!
sizePasses
	| sizer instructions lastForwardJump i limit messageOrLabel |
	sizer := BytecodeSizer assembler: self encoder: encoder.
	"Forward pass that sizes all except forward jumps."
	i := 0.
	limit := (instructions := code instructions) size.
	[(i := i + 1) <= limit] whileTrue:
		[messageOrLabel := instructions at: i.
		sizes
			at: i
			put: ((self isLabel: messageOrLabel)
					ifTrue: [0]
					ifFalse: [index := i.
							(messageOrLabel sentTo: sizer)
								ifNil: [lastForwardJump := i. nil]
								ifNotNil: [:size| size]])].
	lastForwardJump ifNil: [^self].
	"Backward pass that sizes forward jumps."
	i := lastForwardJump + 1.
	[(i := i - 1) >= 1] whileTrue:
		[(sizes at: i) ifNil:
			[index := i.
			 sizes
				at: i
				put: ((instructions at: i) sentTo: sizer)]]! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 3/24/2015 15:54'!
spanToBlockLabel: label "<Association>"
	"N.B. Block labels are expected to follow the last bytecode
	 in the block and preceed the bytecode following the block."
	| labelIndex |
	labelIndex := label value.
	labelIndex ifNil:
		[labelIndex := code instructions identityIndexOf: label.
		 label value: labelIndex].
	self assert: (labelIndex > 1
				and: [(self isLabel: (code instructions at: labelIndex - 1)) not
				and: [#(blockReturnTop blockReturnConstant:
						methodReturnTop methodReturnConstant: methodReturnReceiver)
							includes: (code instructions at: labelIndex - 1) selector]]).
	^(sizes at: labelIndex) ifNotNil:
		[self totalSizeFrom: index + 1 to: labelIndex - 1]! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 14:33'!
spanToLabel: label "<Association>"
	| labelIndex |
	labelIndex := label value.
	labelIndex ifNil:
		[labelIndex := code instructions identityIndexOf: label.
		 label value: labelIndex].
	^(sizes at: labelIndex) ifNotNil:
		[self totalSizeFrom: index + 1 to: labelIndex]! !

!BytecodeAssembler methodsFor: 'assembly' stamp: 'eem 2/22/2013 17:05'!
totalSizeFrom: startPC to: targetPC
	| sum |
	sum := 0.
	^(sizes at: targetPC) ifNotNil:
		[startPC <= targetPC
			ifTrue:
				[startPC to: targetPC do:
					[:i| sum := sum + (sizes at: i)]]
			ifFalse:
				[targetPC to: startPC - 1 do:
					[:i|
					"If a nil is found it will be for a forward branch in the first pass, and
					 the total size will be being requested by a backward branch.  Simply
					 answer anything negative.  All negative branches are long."
					(sizes at: i)
						ifNil: [^-1]
						ifNotNil: [:size| sum := sum - size]]].
		 sum]! !

!BytecodeAssembler methodsFor: 'private' stamp: 'eem 3/24/2017 08:50'!
verify: assemblerMethod
	"If the assemblerMethod contains verification info, check it against the method."

	assemblerMethod header ifNotNil:
		[:expectedHeader|
		self assert: expectedHeader equals: method header entity: 'header'].
	assemblerMethod initialPC ifNotNil:
		[:expectedInitialPC| | alternatePCdelta | "the 32-bit pc if 64-bits or vice verce"
		alternatePCdelta := Smalltalk wordSize = 8 ifTrue: [expectedInitialPC - 1] ifFalse: [(expectedInitialPC - 1) negated].
		self assert: expectedInitialPC
			or: expectedInitialPC + alternatePCdelta
			equals: method initialPC
			entity: 'initialPC'.
		self assert: expectedInitialPC
			or: expectedInitialPC + alternatePCdelta
			equals: method initialPC
			entity: 'endPC']! !

Object subclass: #BytecodeCoverer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Coverage'!

!BytecodeCoverer class methodsFor: 'utilities' stamp: 'eem 3/20/2013 17:33'!
unusedBytecodeSendsEnabled
	^(Smalltalk specialObjectsArray at: 58 ifAbsent: []) == #unusedBytecode! !

!BytecodeCoverer methodsFor: 'private' stamp: 'eem 3/20/2013 17:33'!
checkUnusedBytecodeSendsEnabled
	self class unusedBytecodeSendsEnabled ifFalse:
		[self error: 'unusedBytecode sends are not enabled.  continuting will likely crash the VM']! !

!BytecodeCoverer methodsFor: 'covering' stamp: 'eem 3/20/2013 17:30'!
coverMethod: aMethod
	| labels basicBlockStarts unusedBytecode |
	self checkUnusedBytecodeSendsEnabled.
	labels := (BasicBlockLabellingDisassembler new disassemble: aMethod) labels.
	basicBlockStarts := Dictionary new.
	unusedBytecode := aMethod encoderClass unusedBytecode.
	self assert: unusedBytecode notNil.
	labels do:
		[:assoc| | pc |
		pc := assoc value.
		basicBlockStarts at: pc put: (aMethod at: pc).
		aMethod at: pc put: unusedBytecode].
	aMethod
		propertyValueAt: #coverage put: basicBlockStarts;
		voidCogVMState! !

!BytecodeCoverer methodsFor: 'covering' stamp: 'eem 3/20/2013 17:47'!
uncoverMethod: aMethod
	| coverage |
	coverage := aMethod propertyValueAt: #coverage.
	coverage ifNil: [^self].
	coverage keysAndValuesDo:
		[:pc :bytecode|
		aMethod at: pc put: bytecode].
	aMethod removeProperty: #coverage.
	(aMethod penultimateLiteral isMethodProperties
	 and: [aMethod penultimateLiteral isEmpty]) ifTrue:
		[aMethod penultimateLiteral: aMethod selector]! !

Object subclass: #BytecodeDisassembler
	instanceVariableNames: 'instructions labels labelling instrs thePC method'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Kernel'!

BytecodeDisassembler subclass: #BasicBlockLabellingDisassembler
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Coverage'!

!BasicBlockLabellingDisassembler methodsFor: 'instruction decoding' stamp: 'eem 9/23/2014 15:07'!
jump: offset if: condition
	"Override to add a label for the start of the basic block that follows the jump."
	super jump: offset if: condition.
	labelling ifTrue:
		[labels at: instrs pc put: true]! !

!BasicBlockLabellingDisassembler methodsFor: 'disassembly' stamp: 'eem 3/20/2013 14:02'!
label
	"Override to add a label for the first bytecode pc.
	 Be careful to skip any initial callPrimitive bytecode."
	method isQuick ifFalse:
		[| startPC primCode|
		 startPC := (method primitive ~= 0
					 and: [(primCode := method encoderClass callPrimitiveCode) notNil])
						ifTrue: [method initialPC + ((InstructionStream on: method) bytecodeSize: primCode)]
						ifFalse: [method initialPC].
		 labels at: startPC put: true].
	super label! !

!BasicBlockLabellingDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/20/2013 13:36'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	"Override to add a label for the start of a block."
	labelling ifTrue:
		[labels at: instrs pc put: true].
	super pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 17:34'!
branchIfInstanceOf: aLiteral distance: distance
	(self labelOrNilIfLabellingFor: distance) ifNotNil:
		[:label| | litIndex |
		litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
		litIndex ifNil: [self error: 'could not decode literal index'].
		instructions
			at: thePC
			put: (Message
					selector: #branchIfInstanceOfLiteralAtIndex:distance:
					arguments: {litIndex. label})]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 17:33'!
branchIfNotInstanceOf: aLiteral distance: distance
	(self labelOrNilIfLabellingFor: distance) ifNotNil:
		[:label| | litIndex |
		litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
		litIndex ifNil: [self error: 'could not decode literal index'].
		instructions
			at: thePC
			put: (Message
					selector: #branchIfNotInstanceOfLiteralAtIndex:distance:
					arguments: {litIndex. label})]! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 3/24/2017 08:28'!
disassemble: aMethod
	| endPC |
	method := aMethod.
	labels := Array new: aMethod size + 1 withAll: false.
	instructions := Array new: aMethod size + 1.
	instrs := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	labelling := true.
	[(thePC := instrs pc) <= endPC] whileTrue:
		[instrs interpretNextInstructionFor: self].
	self label.
	labelling := false.
	instrs pc: aMethod initialPC.
	[(thePC := instrs pc) <= endPC] whileTrue:
		[instrs interpretNextInstructionFor: self].
	^(AssemblerMethod
		literals: (self disassembleLiterals: aMethod literals)
		instructions: self instructionSequence
		numArgs: aMethod numArgs
		numTemps: aMethod numTemps
		frameSize: aMethod frameSize
		primitive: aMethod primitive
		flag: aMethod flag
		signFlag: aMethod signFlag
		accessModifier: ([method accessModifierBits] on: MessageNotUnderstood do: [:ex| nil])
		noCounters: ([method hasNoCountersFlag] on: MessageNotUnderstood do: [:ex| [method hasNoCounterBit] on: MessageNotUnderstood do: [:ex2| nil]])
		trailer: aMethod trailer
		methodClass: aMethod methodClass
		selector: aMethod selector
		compiledMethodClass: aMethod class)
			recordVerificationInfoFrom: aMethod;
			yourself
	
	"BytecodeDisassembler new disassemble: BytecodeDisassembler >> #jump:if:"! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'cb 3/20/2017 14:10'!
disassembleLiterals: literals
	literals allButLast withIndexDo: [ :lit :index |
		lit class == CompiledBlock 
			ifTrue: [ literals at: index put: (BytecodeDisassembler new disassemble: lit) ] ].
	^ literals! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 3/24/2015 15:33'!
doesNotUnderstand: aMessage
	(BytecodeSizer includesSelector: aMessage selector) ifFalse:
		[^super doesNotUnderstand: aMessage].
	aMessage lookupClass: nil.
	instructions at: thePC put: aMessage! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 2/22/2013 04:35'!
instructionSequence
	^Array streamContents:
		[:s|
		labels with: instructions do:
			[:label :inst|
			inst ifNotNil:
				[label ~~ false ifTrue:
					[s nextPut: label].
				s nextPut: inst]]]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 17:31'!
jump: offset
	"Print the Unconditional Jump bytecode."

	(self labelOrNilIfLabellingFor: offset) ifNotNil:
		[:label|
		instructions
			at: thePC
			put: (Message
					selector: #jump:
					argument: label)]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 17:31'!
jump: offset if: condition
	"Disassemble the Conditional Jump bytecode."

	(self labelOrNilIfLabellingFor: offset) ifNotNil:
		[:label|
		instructions
			at: thePC
			put: (Message
					selector: #jump:if:
					arguments: { label. condition })]! !

!BytecodeDisassembler methodsFor: 'disassembly' stamp: 'eem 3/20/2013 12:41'!
label
	"N.B. A Label is an Association so that later its pc can be sorted in its value field."
	| n |
	n := 0.
	labels doWithIndex:
		[:l :i|
		l ifTrue:
			[labels at: i put: (Association new key: 'L', (n := n + 1) printString value: i)]]! !

!BytecodeDisassembler methodsFor: 'private' stamp: 'eem 3/23/2017 17:29'!
labelOrNilIfLabellingFor: offset
	"If labelling, just set a flag at the target and answer nil.
	 If done labelling answer the relevant label."
	^labelling
		ifTrue: [labels at: instrs pc + offset put: true. nil]
		ifFalse: [labels at: instrs pc + offset]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:58'!
popIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #popIntoLiteralVariableAtIndex:
				argument: litIndex)! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'ClementBera 3/22/2017 15:39'!
popIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex 
	instructions
		at: thePC
		put: (tempVectorIndex >= 128
				ifTrue:
					[Message
						selector: #popIntoRemoteInstanceVariable:inObjectAt:
						arguments: { remoteTempIndex. tempVectorIndex - 128  }]
				ifFalse:
					[Message
						selector: #popIntoRemoteTemp:inVectorAt:
						arguments: { remoteTempIndex. tempVectorIndex  }])! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 2/22/2013 04:42'!
pushClosureCopyNumCopiedValues: numCopied numArgs: numArgs blockSize: blockSize
	labelling
		ifTrue:
			[labels at: instrs pc + blockSize put: true]
		ifFalse:
			[instructions
				at: thePC
				put: (Message
						selector: #pushClosureCopyNumCopiedValues:numArgs:blockSize:
						arguments: { numCopied. numArgs. labels at: instrs pc + blockSize })]! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:40'!
pushConstant: aLiteral
	instructions
		at: thePC
		put: ((method encoderClass literalIndexOfBytecodeAt: thePC in: method)
				ifNil: [Message
						selector: #pushSpecialConstant:
						argument: aLiteral]
				ifNotNil:
					[:index|
					Message
						selector: #pushConstantAtIndex:
						argument: index])! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 16:51'!
pushFullClosure: compiledBlock numCopied: numCopied
	| litIndex |
	litIndex := method literals indexOf: compiledBlock.
	litIndex ifNil: [self error: 'cannot encode the method'].
	instructions 
		at: thePC 
		put: (Message
				selector: #pushFullClosureAtIndex:numCopied:
				arguments: {litIndex. numCopied})! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/23/2017 16:51'!
pushFullClosure: compiledBlock numCopied: numCopied receiverOnStack: rcvrOnStack ignoreOuterContext: ignoreOuterContext
	| litIndex |
	litIndex := method literals indexOf: compiledBlock.
	litIndex ifNil: [self error: 'cannot encode the method'].
	instructions 
		at: thePC 
		put: (Message
				selector: #pushFullClosureAtIndex:numCopied:receiverOnStack:ignoreOuterContext:
				arguments: {litIndex. numCopied. rcvrOnStack. ignoreOuterContext})! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:59'!
pushLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #pushLiteralVariableAtIndex:
				argument: litIndex)! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'ClementBera 3/22/2017 15:36'!
pushRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex 
	instructions
		at: thePC
		put: (tempVectorIndex >= 128
				ifTrue:
					[Message
						selector: #pushRemoteInstanceVariable:inObjectAt:
						arguments: { remoteTempIndex. tempVectorIndex - 128  }]
				ifFalse:
					[Message
						selector: #pushRemoteTemp:inVectorAt:
						arguments: { remoteTempIndex. tempVectorIndex  }])! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'eem 3/14/2013 16:59'!
storeIntoLiteralVariable: aLiteral
	| litIndex |
	litIndex := method encoderClass literalIndexOfBytecodeAt: thePC in: method.
	litIndex ifNil: [self error: 'could not decode literal index'].
	instructions
		at: thePC
		put: (Message
				selector: #storeIntoLiteralVariableAtIndex:
				argument: litIndex)! !

!BytecodeDisassembler methodsFor: 'instruction decoding' stamp: 'ClementBera 3/22/2017 15:39'!
storeIntoRemoteTemp: remoteTempIndex inVectorAt: tempVectorIndex 
	instructions
		at: thePC
		put: (tempVectorIndex >= 128
				ifTrue:
					[Message
						selector: #storeIntoRemoteInstanceVariable:inObjectAt:
						arguments: { remoteTempIndex. tempVectorIndex - 128  }]
				ifFalse:
					[Message
						selector: #storeIntoRemoteTemp:inVectorAt:
						arguments: { remoteTempIndex. tempVectorIndex  }])! !

!MethodContext methodsFor: '*MethodMassage-coverage' stamp: 'bwesterg 9/30/2013 17:32'!
unusedBytecode
	"Handle unusedBytecode by replacing the bytecode with the
	 correct one found in the coverage property and continuing.
	 Continue via wait/signal since return would push a result."
	| coverage semaphore process originalBytecode |
	self assert: (method at: pc) = method encoderClass unusedBytecode.
	coverage := method propertyValueAt: #coverage.
	self assert: coverage notNil.
	originalBytecode := coverage at: pc ifAbsent: [nil].
	self assert: originalBytecode notNil.
	semaphore := Semaphore new.
	process := Processor activeProcess.

	[method at: pc put: originalBytecode.
	 coverage removeKey: pc ifAbsent: [].
	 process suspendedContext unwindTo: self.
	 process suspendedContext: self.
	 semaphore signal] fork.

	semaphore wait! !

!CompiledMethodTrailer methodsFor: '*MethodMassage-raw data' stamp: 'eem 3/15/2013 10:02'!
rawData
	^(ByteArray new: size)
		replaceFrom: 1
		to: size
		with: method
		startingAt: method size - size + 1! !

TestCase subclass: #BytecodeAssemblerTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Tests'!

!BytecodeAssemblerTests class methodsFor: 'utilities' stamp: 'eem 3/15/2013 13:04'!
fixupMethodClassBindings
	"Work-around the Monticello bug that results in not sharing the binding in Smalltalk as a method's methodClassAssociation."
	"self fixupMethodClassBindings"
	self systemNavigation allSelect:
		[:m|
		(m methodClass isMeta not
		 and: [m methodClassAssociation key isNil]) ifTrue:
			[m methodClassAssociation: m methodClass binding].
		false]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/23/2013 16:08'!
contextPartMethodWithMostOf: binaryBlock
	^(Array streamContents: [:s| ContextPart withAllSubclassesDo: [:sc| s nextPutAll: sc methodDict values]])
		inject: ContextPart methodDict anyOne
		into:
			[:m1 :m2|
			((1 to: m1 methodClass instSize) select: [:i| binaryBlock value: m1 value: i]) size >=
			((1 to: m2 methodClass instSize) select: [:i| binaryBlock value: m2 value: i]) size
				ifTrue: [m1]
				ifFalse: [m2]]! !

!BytecodeAssemblerTests methodsFor: 'accessing' stamp: 'eem 2/23/2013 23:12'!
defaultTimeout
	"Answer the default timeout to use for tests in this test case.
	The timeout is a value in seconds."

	^30 "seconds"! !

!BytecodeAssemblerTests methodsFor: 'testing' stamp: 'eem 3/24/2015 16:31'!
expectedFailures
	((thisContext method respondsTo: #usesAlternateBytecodeSet)
	 or: [thisContext method respondsTo: #bytecodeSetName]) ifTrue:
		[^#()].
	^#(testDuplicateNewsqueakV4Methods)! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 3/20/2013 12:48'!
testAsAssemblerFor: aMethod
	"self new testAsAssemblerFor: self >> #testAsAssemblerFor:"
	| assemblerMethod assembler reassembledMethod |
	assemblerMethod := BytecodeDisassembler new disassemble: aMethod.
	assembler := String streamContents: [:s| assemblerMethod printAsAssemblerOn: s].
	self shouldnt: [reassembledMethod := Compiler evaluate: assembler]
		raise: Error.
	self assert: assemblerMethod compiledMethodClass equals: reassembledMethod compiledMethodClass.
	self assert: assemblerMethod flag equals: reassembledMethod flag.
	self assert: assemblerMethod frameSize equals: reassembledMethod frameSize.
	self assert: assemblerMethod methodClass equals: reassembledMethod methodClass.
	self assert: assemblerMethod numArgs equals: reassembledMethod numArgs.
	self assert: assemblerMethod numTemps equals: reassembledMethod numTemps.
	self assert: assemblerMethod primitive equals: reassembledMethod primitive.
	self assert: assemblerMethod selector equals: reassembledMethod selector.
	self assert: assemblerMethod signFlag equals: reassembledMethod signFlag.
	1 to: reassembledMethod literals size do:
		[:i|
		self assert: (assemblerMethod literals at: i)
			equals: (reassembledMethod literals at: i)].
	assemblerMethod instructions
		ifNil: [self assert: reassembledMethod instructions isNil]
		ifNotNil:
			[:insts|
			 self assert: insts size equals: reassembledMethod instructions size.
			 "Nil the labels' values for analogousCodeTo: comparison below; the values
			  are usefully the bytecode pcs of the labels, but this fails the comparisons."
			 insts do: [:inst| inst isVariableBinding ifTrue: [inst value: nil]].
			 insts
				with: reassembledMethod instructions
				do: [:old :new|
					old isVariableBinding
						ifTrue: [self assert: old key = new key]
						ifFalse: [self assert: (old analogousCodeTo: new)]]].
	self assert: aMethod equals: (BytecodeAssembler new assemble: reassembledMethod)! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/15/2013 08:57'!
testAsAssemblerForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testAsAssemblerFor: m]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 9/23/2014 16:17'!
testBasicBlockLabelsFor: aMethod
	"self new testBasicBlockLabelsFor: self >> #testBasicBlockLabelsFor:"
	| labels stream endPC nextPC |
	labels := (BasicBlockLabellingDisassembler new disassemble: aMethod) labels.
	aMethod isQuick ifTrue:
		[^self assert: labels isEmpty].
	self assert: (labels select: [:l| aMethod initialPC = l value]) size = 1.
	aMethod embeddedBlockClosures do:
		[:closure|
		self assert: (labels select: [:l| closure startpc = l value]) size = 1].
	stream := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	[stream pc <= endPC] whileTrue:
		[nextPC := stream nextPc: (aMethod at: stream pc).
		 (stream willJumpIfTrue or: [stream willJumpIfFalse]) ifTrue:
			["A conditional jump should have a label following it."
			 self assert: (labels select: [:l| stream followingPc = l value]) size = 1].
		 stream interpretJump ifNotNil:
			[:dist| "There shoud be a label at the target, and, if it is conditional, also one following the jump."
			self assert: (labels select: [:l| stream pc + dist = l value]) size = 1].
		 stream pc: nextPC]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/20/2013 13:42'!
testBasicBlockLabelsForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testBasicBlockLabelsFor: m]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/22/2013 11:35'!
testDisassembleBytecodeAssemblerMethods
	self testDisassembleMethodsFor: BytecodeAssembler! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/22/2013 11:38'!
testDisassembleMethodsFor: aBehavior
	aBehavior selectors asArray sort do:
		[:sel| | method |
		method := aBehavior compiledMethodAt: sel.
		self shouldnt: [BytecodeDisassembler new disassemble: method]
			raise: Error]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/23/2013 15:49'!
testDuplicateAllInstructionStreamMethods
	InstructionStream withAllSuperclasses, InstructionStream allSubclasses do:
		[:isc|
		self testDuplicateMethodsFor: isc
			if: [:method| true]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 2/23/2013 15:40'!
testDuplicateBytecodeAssemblerMethods
	self testDuplicateMethodsFor: BytecodeAssembler
		if: [:method| true]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 2/23/2013 17:54'!
testDuplicateMethodsFor: aBehavior if: methodSelectBlock
	aBehavior selectors asArray sort do:
		[:sel| | method disassembly assembly |
		method := aBehavior compiledMethodAt: sel.
		(methodSelectBlock value: method) ifTrue:
			[disassembly := BytecodeDisassembler new disassemble: method.
			 assembly := BytecodeAssembler new assemble: disassembly.
			"method indexOfDifferenceWith: assembly"
			 self assert: method equals: assembly]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/24/2015 16:34'!
testDuplicateNewsqueakV4Methods
	Smalltalk allClasses do:
		((thisContext method respondsTo: #usesAlternateBytecodeSet)
			ifTrue:
				[[:class|
				  self testDuplicateMethodsFor: class
					if: [:method| method usesAlternateBytecodeSet].
				  self testDuplicateMethodsFor: class class
					if: [:method| method usesAlternateBytecodeSet]]]
			ifFalse:
				[[:class|
				  self testDuplicateMethodsFor: class
					if: [:method| method bytecodeSetName = 'NewsqueakV4'].
				  self testDuplicateMethodsFor: class class
					if: [:method| method bytecodeSetName = 'NewsqueakV4']]])! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 4/9/2013 12:13'!
testDuplicateSqueakV3Methods
	"AndreasProfiler spyOn: [self new testDuplicateSqueakV3Methods]"
	| filter |
	filter := (thisContext method respondsTo: #usesAlternateBytecodeSet)
				ifTrue: [[:method| method usesAlternateBytecodeSet not]]
				ifFalse: [[:method| true]].
	Smalltalk allClasses do:
		[:class|
		self testDuplicateMethodsFor: class if: filter.
		self testDuplicateMethodsFor: class class if: filter]! !

!BytecodeAssemblerTests methodsFor: 'test support' stamp: 'eem 3/20/2013 13:15'!
testLabelsFor: aMethod
	"self new testLabelsFor: self >> #testLabelsFor:"
	| labels stream endPC |
	labels := (BytecodeDisassembler new disassemble: aMethod) labels.
	stream := InstructionStream on: aMethod.
	endPC := aMethod endPC.
	[stream pc <= endPC] whileTrue:
		[stream interpretJump
			ifNotNil: [:dist| self assert: (labels select: [:l| stream pc + dist = l value]) size = 1]
			ifNil: [stream pc: (stream nextPc: (aMethod at: stream pc))]]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 3/20/2013 13:17'!
testLabelsForThisPackage
	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testLabelsFor: m]! !

!BytecodeAssemblerTests methodsFor: 'tests' stamp: 'eem 8/7/2014 11:51'!
testManualAssembly
	self
		assert:
			(AssemblerMethod new
				methodClass: Object; selector: #printOn:; numArgs: 1; numTemps: 2;
				pushReceiver;
				send: #class;
				send: #name;
				popIntoTemporaryVariable: 1;
				pushTemporaryVariable: 0;
				doDup;
				pushTemporaryVariable: 1;
				send: #first;
				send: #isVowel;
				jump: 'L1' if: false;
				pushConstant: 'an ';
				jump: 'L2';
			label: 'L1';
				pushConstant: 'a ';
			label: 'L2';
				send: #nextPutAll:;
				doPop;
				pushTemporaryVariable: 1;
				send: #nextPutAll:;
				doPop;
				methodReturnReceiver;
				assemble) decompile printString
		equals: (Object >>#printOn:) decompile printString! !

TestCase subclass: #BytecodeCoverageTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MethodMassage-Tests'!

!BytecodeCoverageTests methodsFor: 'tests support' stamp: 'eem 3/21/2013 14:45'!
coverTestBasicBlockLabelsForMethod

	(PackageOrganizer default packageOfClass: self class) actualMethodsDo:
		[:m| self testBasicBlockLabelsFor: m]! !

!BytecodeCoverageTests methodsFor: 'initialize-release' stamp: 'eem 3/20/2013 16:11'!
setUp
	self class recompile: #testBasicBlockLabelsFor: from: BytecodeAssemblerTests! !

!BytecodeCoverageTests methodsFor: 'initialize-release' stamp: 'eem 3/20/2013 16:13'!
tearDown
	self class removeSelectorSilently: #testBasicBlockLabelsFor:! !

!BytecodeCoverageTests methodsFor: 'tests' stamp: 'eem 9/23/2014 15:38'!
testCoverageOfTestBasicBlockLabelsForMethod
	"[| me |
	me := self new.
	[me setUp; testCoverageOfTestBasicBlockLabelsForMethod]
		ensure: [me tearDown]] timeToRun"
	"[| me |
	me := self new.
	[me setUp; coverTestBasicBlockLabelsForMethod]
		ensure: [me tearDown]] timeToRun"
	| theMethod |
	theMethod := self class >> #testBasicBlockLabelsFor:.
	BytecodeCoverer new coverMethod: theMethod.
	self assert: (theMethod propertyValueAt: #coverage) notNil.
	self assert: (theMethod propertyValueAt: #coverage) notEmpty.

	self coverTestBasicBlockLabelsForMethod.

	self assert: (theMethod propertyValueAt: #coverage) isEmpty! !

!BytecodeEncoder class methodsFor: '*MethodMassage-assembly' stamp: 'eem 8/7/2014 11:13'!
assembler
	"Answer an assembler for a method that uses the receiver's instruction set."
	^AssemblerMethod new
		encoder: self new;
		yourself! !

!BytecodeEncoder methodsFor: '*MethodMassage-literal access' stamp: 'eem 2/23/2013 23:04'!
litIndexOrNilFor: aLiteral
	| index |
	index := aLiteral isVariableBinding
				ifTrue: [literalStream originalContents identityIndexOf: aLiteral ifAbsent: [^nil]]
				ifFalse: [literalStream originalContents indexOf: aLiteral ifAbsent: [^nil]].
	^index <= literalStream position ifTrue:
		[index - 1]! !

!EncoderForV3 class methodsFor: '*MethodMassage-disassembly' stamp: 'eem 5/19/2014 17:21'!
literalIndexOfBytecodeAt: pc in: aCompiledMethod
	"32-63 		001iiiii 						Push Literal Constant #iiiii
	 64-95 		010iiiii 						Push Literal Variable #iiiii
	 128 		10000000 jjkkkkkk 		Push (Receiver Variable, Temporary Location, Literal Constant, Literal Variable) [jj] #kkkkkk
	129 		10000001 jjkkkkkk 		Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk
	130 		10000010 jjkkkkkk 		Pop and Store (Receiver Variable, Temporary Location, Illegal, Literal Variable) [jj] #kkkkkk
	132 		10000100 iiijjjjj kkkkkkkk 	(Send, Send Super, Push Receiver Variable, Push Literal Constant, Push Literal Variable, Store Receiver Variable, Store-Pop Receiver Variable, Store Literal Variable)[iii] #kkkkkkkk jjjjj (for sends jjjjj = numArgs)"
	| byte type |
	byte := aCompiledMethod at: pc.
	byte <= 95 ifTrue:
		[^byte >= 32 ifTrue:
			[byte bitAnd: 16r1F]].
	(byte between: 128 and: 130) ifTrue:
		[^(aCompiledMethod at: pc + 1) bitAnd: 16r3F].
	byte = 132 ifTrue:
	[type := (aCompiledMethod at: pc + 1) bitShift: -5.
	 (type = 3 or: [type = 4 or: [type = 7]]) ifTrue:
		[^aCompiledMethod at: pc + 2]].
	^nil! !

!CompiledMethod methodsFor: '*MethodMassage-disassembly' stamp: 'eem 4/6/2015 11:35'!
assemblerString
	^self disassemble assemblerString! !

!CompiledMethod methodsFor: '*MethodMassage-disassembly' stamp: 'eem 3/15/2013 13:16'!
disassemble
	^BytecodeDisassembler new disassemble: self! !

!CompiledMethod methodsFor: '*MethodMassage-debugging' stamp: 'eem 2/23/2013 16:30'!
indexOfDifferenceWith: method
	"Answer whether the receiver implements the same code as the argument, method.
	 Here ``same code'' means that if the receiver's source is compiled with the same
	 compiler it should produce the same sequence of bytecodes and literals, same
	 trailer and same properties.  Hence this definition of #= (only one of many plausible
	 definitions) can be used to quickly identify changes in the compiler's output."
	| numLits |
	method isCompiledMethod ifFalse: [^1].
	self size = method size ifFalse: [^-2].
	self header = method header ifFalse: [^-3]. "N.B. includes numLiterals comparison."
	self initialPC to: self endPC do:
		[:i | (self at: i) = (method at: i) ifFalse: [^i]].
	numLits := self numLiterals.
	1 to: numLits do:
		[:i| | lit1 lit2 |
		lit1 := self literalAt: i.
		lit2 := method literalAt: i.
		(lit1 == lit2 or: [lit1 literalEqual: lit2]) ifFalse:
			[(i = 1 and: [#(117 120) includes: self primitive])
				ifTrue:
					[lit1 isArray
						ifTrue:
							[(lit2 isArray and: [lit1 allButLast = lit2 allButLast]) ifFalse:
								[^i]]
						ifFalse: "ExternalLibraryFunction"
							[(lit1 analogousCodeTo: lit2) ifFalse:
								[^i]]]
				ifFalse:
					[i = (numLits - 1)
						ifTrue: "properties"
							[(self properties analogousCodeTo: method properties)
								ifFalse: [^i]]
						ifFalse: "last literal (methodClassAssociation) of class-side methods is not unique"
							[(i = numLits
							 and: [lit1 isVariableBinding
							 and: [lit2 isVariableBinding
							 and: [lit1 key == lit2 key
							 and: [lit1 value == lit2 value]]]]) ifFalse:
								[^i]]]]].
	^0! !

!Encoder methodsFor: '*MethodMassage-instance initialization' stamp: 'eem 8/7/2014 11:44'!
initClass: targetBehavior literalStream: literalsOrNil
	cue := CompilationCue class: targetBehavior.
	class := targetBehavior.
	optimizedSelectors := Set new.
	literalsOrNil
		ifNil:
			[literalStream := WriteStream on: (Array new: 32).
			 addedSelectorAndMethodClassLiterals := false]
		ifNotNil:
			[literalStream := (WriteStream on: literalsOrNil asArray from: 1 to: literalsOrNil size) setToEnd.
			 addedSelectorAndMethodClassLiterals := true]! !

!EncoderForSistaV1 class methodsFor: '*MethodMassage-disassembly' stamp: 'eem 3/24/2017 09:34'!
literalIndexOfBytecodeAt: pc in: method
	"16-31		0001 iiii						Push Literal Variable #iiii
	 32-63		001 iiiii							Push Literal #iiiii
*	 227		11100011	iiiiiiii				Push Literal Variable #iiiiiiii (+ Extend A * 256)
*	 228		11100100	iiiiiiii				Push Literal #iiiiiiii (+ Extend A * 256)
**	 241	(3)	11110001	iiiiiiii				Pop and Store Literal Variable #iiiiiiii (+ Extend A * 256)
**	 244	(3)	11110100	iiiiiiii				Store Literal Variable #iiiiiiii (+ Extend A * 256)
**	 254		11111110	kkkkkkkk	jjjjjjjj	Branch If Not Instance Of Behavior/Array Of Behavior literal kkkkkkkk (+ Extend A * 256, where Extend A >= 0) distance jjjjjjjj (+ Extend B * 256, where Extend B >= 0 and <= 127)
**	254		11111110	kkkkkkkk	jjjjjjjj	Branch If Instance Of Behavior/Array Of Behavior literal kkkkkkkk (+ Extend A * 256, where Extend A >= 0) distance jjjjjjjj (+ (Extend B bitAnd: 127) * 256, where Extend B >= 128 and <= 255)
"
	^self extensionsAt: pc in: method into:
		[:extA :extB :nExts| | byte |
		byte := method at: pc + nExts.
		byte < 64
			ifTrue:
				[byte < 32
					ifTrue: [byte > 15 ifTrue: [byte bitAnd: 16rF]]
					ifFalse: [byte bitAnd: 16r1F]]
			ifFalse:
				[(byte >= 227 and: [byte <= 228 or: [#[241 244 254] includes: byte]]) ifTrue:
					[extA * 256 + (method at: pc + nExts + 1)]]]! !
AssemblerEncoderInterface initialize!
